from typing import Tuple

import numpy as np
import statsmodels.api as sm


def evaluate_density(samples, points, kernel, bandwidth) -> Tuple[np.ndarray, np.ndarray]:
    """For a given dataset, computes the kernel densities at the given points,
    and returns back the points and density arrays."""

    # By default, we'll use a 'hard' KDE span. That is, we'll
    # evaluate the densities and N equally spaced points
    # over the range [min(samples), max(samples)]
    if isinstance(points, int):
        points = np.linspace(np.min(samples), np.max(samples), points)

    # Unless a specific range is specified...
    else:
        points = np.asarray(points)

    if points.ndim > 1:
        raise ValueError(
            f"The 'points' at which KDE is computed should be represented by a "
            f"one-dimensional array, got an array of shape {points.shape} instead."
        )

    # I decided to use statsmodels' KDEUnivariate for KDE. There are many
    # other supported alternatives in the python scientific computing
    # ecosystem. See, for instance, scipy's alternative - on which
    # statsmodels relies - `from scipy.stats import gaussian_kde`
    dens = sm.nonparametric.KDEUnivariate(samples)

    # I'm hard-coding the `fft=self.kernel == "gau"` for convenience here.
    # This avoids the need to expose yet another __init__ argument (fft)
    # to this class. The drawback is that, if and when statsmodels
    # implements another kernel with fft, this will fallback to
    # using the unoptimised version (with fft = False).
    dens.fit(kernel=kernel, fft=kernel == "gau", bw=bandwidth)
    densities = dens.evaluate(points)

    # I haven't investigated the root of this issue yet
    # but statsmodels' KDEUnivariate implementation
    # can return a nan float if something goes
    # wrong internally. As to avoid confusion
    # further down the pipeline, I decided
    # to check whether the correct object
    # (and shape) are being returned.
    if not isinstance(densities, np.ndarray) or densities.shape != points.shape:
        raise RuntimeError(
            f"Could now evaluate densities using the {repr(kernel)} kernel! " f"Try using kernel='gau' (default)."
        )

    return points, densities


def get_densities(samples, points, kernel, bandwidth) -> np.ndarray:
    return np.asarray([evaluate_density(samples=s, points=points, kernel=kernel, bandwidth=bandwidth) for s in samples])



from typing import Callable, Dict, List, Optional

import numpy as np
from _plotly_utils.colors import validate_colorscale
from plotly import graph_objects as go

from ridgeplot._colors import ColorScaleType, apply_alpha, get_color, get_plotly_colorscale
from ridgeplot._utils import get_extrema_3d, normalise


class _RidgePlotFigureFactory:
    """Refer to ridgeplot.ridgeplot() for docstring."""

    def __init__(
        self,
        densities,
        colorscale,
        coloralpha,
        colormode,
        labels,
        linewidth,
        spacing,
        show_annotations,
        xpad,
    ) -> None:
        # ==============================================================
        # ---  Get clean and validated input arguments
        # ==============================================================
        n_traces = len(densities)

        # Check whether all density arrays have shape (2, N)
        new_densities = []
        for i in range(n_traces):
            d = np.asarray(densities[i])
            if d.shape[0] != 2 or d.ndim != 2:
                raise ValueError(f"Each density array should have shape (2, N), got {d.shape}")
            new_densities.append(d)

        if isinstance(colorscale, str):
            colorscale = get_plotly_colorscale(name=colorscale)
        validate_colorscale(colorscale)

        if colormode not in self.colormode_maps.keys():
            raise ValueError(
                f"The colormode argument should be one of "
                f"{tuple(self.colormode_maps.keys())}, got {colormode} instead."
            )

        if coloralpha is not None:
            coloralpha = float(coloralpha)

        if labels is not None:
            n_labels = len(labels)
            if n_labels != n_traces:
                raise ValueError(f"Expected {n_traces} labels, got {n_labels}.")
            labels = list(map(str, labels))
        else:
            labels = [f"Trace {i + 1}" for i in range(n_traces)]

        self.densities: List[np.ndarray] = new_densities
        self.colorscale: ColorScaleType = colorscale
        self.coloralpha: Optional[float] = coloralpha
        self.colormode = str(colormode)
        self.labels: list = labels
        self.linewidth: float = float(linewidth)
        self.spacing: float = float(spacing)
        self.show_annotations: bool = bool(show_annotations)
        self.xpad: float = float(xpad)

        # ==============================================================
        # ---  Other instance variables
        # ==============================================================
        self.n_traces: int = n_traces
        self.x_min, self.x_max, _, self.y_max = get_extrema_3d(densities)
        self.fig: go.Figure = go.Figure()
        self.colors: List[str] = self.pre_compute_colors()

    @property
    def colormode_maps(self) -> Dict[str, Callable[[], List[float]]]:
        return {
            "index": self._compute_midpoints_index,
            "mean-minmax": self._compute_midpoints_mean_minmax,
            "mean-means": self._compute_midpoints_mean_means,
        }

    def draw_base(self, x, y_shifted) -> None:
        """Adds an invisible trace at constant y that will serve as the
        fill-limit for the corresponding density trace."""
        self.fig.add_trace(
            go.Scatter(
                x=x,
                y=[y_shifted] * len(x),
                # make trace 'invisible'
                # Note: visible=False does not work with fill="tonexty"
                line=dict(color="rgba(0,0,0,0)", width=0),
                showlegend=False,
            )
        )

    def draw_density_trace(self, x, y, label, color) -> None:
        """Adds a density 'trace' to the Figure. The fill="tonexty" option
        fills the trace until the previously drawn trace (see `draw_base`)."""
        line_color = "rgba(0,0,0,0.6)" if color is not None else None
        self.fig.add_trace(
            go.Scatter(
                x=x,
                y=y,
                fillcolor=color,
                name=label,
                fill="tonexty",
                mode="lines",
                line=dict(color=line_color, width=self.linewidth),
            ),
        )

    def update_layout(self, y_ticks: list) -> None:
        """Update figure's layout."""
        self.fig.update_layout(
            hovermode=False,
            legend=dict(traceorder="normal"),
        )
        axes_common = dict(
            zeroline=False,
            showgrid=True,
        )
        self.fig.update_yaxes(
            showticklabels=self.show_annotations,
            tickvals=y_ticks,
            ticktext=self.labels,
            **axes_common,
        )
        x_padding = self.xpad * (self.x_max - self.x_min)
        self.fig.update_xaxes(
            range=[self.x_min - x_padding, self.x_max + x_padding],
            showticklabels=True,
            **axes_common,
        )

    def _compute_midpoints_index(self) -> List[float]:
        return [i / (self.n_traces - 1) for i in reversed(range(self.n_traces))]

    def _compute_midpoints_mean_minmax(self) -> List[float]:
        means = [np.sum(x * y) / np.sum(y) for x, y in self.densities]
        return [normalise(mean, min_=self.x_min, max_=self.x_max) for mean in means]

    def _compute_midpoints_mean_means(self) -> List[float]:
        means = [np.sum(x * y) / np.sum(y) for x, y in self.densities]
        return [normalise(mean, min_=min(means), max_=max(means)) for mean in means]

    def pre_compute_colors(self) -> List[str]:
        midpoints = self.colormode_maps[self.colormode]()
        colors = []
        for midpoint in midpoints:
            color = get_color(self.colorscale, midpoint=midpoint)
            if self.coloralpha is not None:
                color = apply_alpha(color, alpha=self.coloralpha)
            colors.append(color)
        return colors

    def make_figure(self) -> go.Figure:
        y_ticks = []
        for i, ((x, y), label, color) in enumerate(zip(self.densities, self.labels, self.colors)):
            # y_shifted is the y-origin for the new trace
            y_shifted = -i * (self.y_max * self.spacing)
            self.draw_base(x=x, y_shifted=y_shifted)
            self.draw_density_trace(x=x, y=y + y_shifted, label=label, color=color)
            y_ticks.append(y_shifted)
        self.update_layout(y_ticks=y_ticks)
        return self.fig



import json
from numbers import Number
from pathlib import Path
from typing import Dict, List, Tuple, Union

from plotly.colors import find_intermediate_color, hex_to_rgb, label_rgb

from ridgeplot._utils import normalise

# Ideally: ColorScaleType = List[Tuple[Number, str]]
ColorScaleType = List[List[Union[Number, str]]]

_path_to_colors_dict = Path(__file__).parent.joinpath("colors.json")
with _path_to_colors_dict.open() as _colors_json:
    ALL_PLOTLY_COLORS_SCALES: Dict[str, ColorScaleType] = json.load(_colors_json)


def _any_to_rgb(color: Union[tuple, str]) -> str:
    if isinstance(color, tuple):
        color = label_rgb(color)
    if color.startswith("#"):
        color = label_rgb(hex_to_rgb(color))
    if not color.startswith("rgb("):
        raise RuntimeError("Something went wrong with the logic above!")
    return color


def named_colorscales() -> Tuple[str]:
    return tuple(ALL_PLOTLY_COLORS_SCALES.keys())


def get_plotly_colorscale(name: str) -> ColorScaleType:
    """Helper to get a known Plotly colorscale, raising a ValueError if an
    invalid name is provided."""
    if name not in ALL_PLOTLY_COLORS_SCALES:
        raise ValueError(
            f"Could not find colorscale '{name}'. The available colorscale"
            f" names are {tuple(ALL_PLOTLY_COLORS_SCALES.keys())}."
        )
    return ALL_PLOTLY_COLORS_SCALES[name]


def get_color(colorscale: ColorScaleType, midpoint: float) -> str:
    """Given a colorscale, it interpolates the expected color at a given
    midpoint, on a scale from 0 to 1."""
    if 0 > midpoint > 1:
        raise ValueError(f"The 'midpoint' should be a float value between 0 and 1, not {midpoint}.")
    scale = [s for s, _ in colorscale]
    colors = [_any_to_rgb(c) for _, c in colorscale]
    del colorscale
    if midpoint in scale:
        return colors[scale.index(midpoint)]
    ceil = min(filter(lambda s: s > midpoint, scale))
    floor = max(filter(lambda s: s < midpoint, scale))
    midpoint_normalised = normalise(midpoint, min_=floor, max_=ceil)
    color = find_intermediate_color(
        lowcolor=colors[scale.index(floor)],
        highcolor=colors[scale.index(ceil)],
        intermed=midpoint_normalised,
        colortype="rgb",
    )
    return color


def apply_alpha(color: Union[tuple, str], alpha) -> str:
    color = _any_to_rgb(color)
    return f"rgba({color[4:-1]}, {alpha})"



import numpy as np
import pandas as pd

from ridgeplot import ridgeplot

# Get the raw data
df = pd.read_csv("https://raw.githubusercontent.com/bokeh/bokeh/main/bokeh/sampledata/_data/probly.csv")


# Let's grab only the subset of columns displayed in the example
column_names = [
    "Almost Certainly",
    "Very Good Chance",
    "We Believe",
    "Likely",
    "About Even",
    "Little Chance",
    "Chances Are Slight",
    "Almost No Chance",
]
df = df[column_names]


# Not only does 'ridgeplot(...)' come configured with sensible defaults
# but is also fully configurable to your own style and preference!
fig = ridgeplot(
    # Get your samples in the correct format
    samples=df.values.T,
    # We can specify the bandwidth used for KDE
    bandwidth=4,
    # and the range of points for which the KDE is evaluated
    kde_points=np.linspace(-12.5, 112.5, 400),
    # You can pass any plotly color scale here
    colorscale="viridis",
    # In the example, the distributions seem to be colored
    # by 'index'. Have a look at the other available options!
    colormode="index",
    # Set the transparency level
    coloralpha=0.6,
    # Always label your plots! Dont be evil...
    labels=column_names,
    # Adjust the vertical spacing between the distributions
    spacing=5 / 9,
)

# Again, update the figure layout to your liking here
fig.update_layout(
    title="What probability would you assign to the phrase <i>“Highly likely”</i>?",
    height=650,
    width=800,
    plot_bgcolor="rgba(255, 255, 255, 0.0)",
    xaxis_gridcolor="rgba(0, 0, 0, 0.1)",
    yaxis_gridcolor="rgba(0, 0, 0, 0.1)",
    yaxis_title="Assigned Probability (%)",
)
fig.show()



"""ridgeplot: beautiful ridgeline plots in Python

The ridgeplot python library aims at providing a simple API for plotting
beautiful ridgeline plots within the extensive Plotly interactive graphing
environment.


  Simple example:

  from numpy.random import normal
  from ridgeplot import ridgeplot

  # Put your real samples here...
  synthetic_samples = [normal(n / 1.2, size=600) for n in reversed(range(9))]

  # The `ridgeplot()` helper comes packed with sensible defaults
  fig = ridgeplot(samples=synthetic_samples)

  # and the returned Plotly figure is still fully customizable
  fig.update_layout(height=500, width=800)

  # show us the work!!
  fig.show()

"""
from ridgeplot._colors import named_colorscales
from ridgeplot._ridgeplot import ridgeplot
from ridgeplot._version import __version__

__all__ = ["ridgeplot", "named_colorscales", "__version__"]



from typing import List, Tuple

import numpy as np


def get_extrema_2d(arr) -> Tuple[float, float]:
    """Calculates and returns the extrema (min, max) of a 2D (N, M) array."""
    arr = np.asarray(arr).flat
    return np.min(arr), np.max(arr)


def get_extrema_3d(arr: List[np.ndarray]) -> Tuple[float, float, float, float]:
    """Calculates and returns the x-y extrema (x_min, x_max, y_min, y_max) of
    a 3D (N, 2, M) array."""
    x_min = 0
    x_max = 0
    y_min = 0
    y_max = 0
    for x in arr:
        x_min = min(x_min, np.min(x[0]))
        x_max = max(x_max, np.max(x[0]))
        y_min = min(y_min, np.min(x[1]))
        y_max = max(y_max, np.max(x[1]))
    return x_min, x_max, y_min, y_max


def normalise(val: float, min_: float, max_: float) -> float:
    assert max_ > min_
    return (val - min_) / (max_ - min_)



from numpy.random import normal

from ridgeplot import ridgeplot

# Put your real samples here...
synthetic_samples = [normal(n / 1.2, size=600) for n in reversed(range(9))]

# The `ridgeplot()` helper comes packed with sensible defaults
fig = ridgeplot(samples=synthetic_samples)

# and the returned Plotly figure is still fully customizable
fig.update_layout(height=500, width=800)

# show us the work!!
fig.show()



from typing import Optional, Union

import plotly.graph_objects as go

from ridgeplot._colors import ColorScaleType
from ridgeplot._figure_factory import _RidgePlotFigureFactory
from ridgeplot._kde import get_densities


def ridgeplot(
    samples=None,
    densities=None,
    kernel: str = "gau",
    bandwidth="normal_reference",
    kde_points=500,
    colorscale: Union[str, ColorScaleType] = "plasma",
    colormode: str = "mean-means",
    coloralpha: Optional[float] = None,
    labels=None,
    linewidth: float = 1.4,
    spacing: float = 0.5,
    show_annotations: bool = True,
    xpad: float = 0.05,
) -> go.Figure:
    """Creates and returns a Plotly figure with a beautiful ridgeline plot.

    Note:
      You have to specify one of: `samples` or `densities`. If you specify
      both `samples` and `densities` arguments, a `ValueError` exception will
      be raised! One of these arguments should always remain set to `None`.
      See `samples` and `densities` bellow for more information.

    Args:
      samples:
        If `samples` data is specified, Kernel Density Estimation (KDE) will be
        computed. See `kernel`, `bandwidth`, and `kde_points` for more details
         and KDE configuration options.
      densities:
        If `densities` arrays are specified instead, the KDE step will be
        skipped and all associated arguments ignored. Each density array should
        have shape `(2, N)`, but `N` may vary with each array.
      kernel:
        The Kernel to be used during Kernel Density Estimation. The default is
        a Gaussian Kernel ("gau"). Choices are:
          - "biw" for biweight
          - "cos" for cosine
          - "epa" for Epanechnikov
          - "gau" for Gaussian.
          - "tri" for triangular
          - "triw" for triweight
          - "uni" for uniform
      bandwidth:
        The bandwidth to use during Kernel Density Estimation. The default is
        "normal_reference". Choices are:
          - "scott" - 1.059 * A * nobs ** (-1/5.), where A is
          `min(std(x),IQR/1.34)`
          - "silverman" - .9 * A * nobs ** (-1/5.), where A is
            `min(std(x),IQR/1.34)`
          - "normal_reference" - C * A * nobs ** (-1/5.), where C is calculated
           from the kernel. Equivalent (up to 2 dp) to the "scott" bandwidth
           for gaussian kernels. See `statsmodels/nonparametric/bandwidths.py`
          - If a float is given, its value is used as the bandwidth.
          - If a callable is given, it's return value is used. The callable
          should take exactly two parameters, i.e., fn(x, kern), and return a
          float, where:
            * x - the clipped input data
            * kern - the kernel instance used
      kde_points:
        This argument controls the points at which KDE is computed. If an `int`
        value is passed (default), the densities will be evaluated at
        `kde_points` evenly spaced points between the min and max of each set
        of samples. However, you may also specify a custom range by instead
        passing an array of points. This array should be one-dimensional.
      colorscale:
        Any valid Plotly color-scale or a `str` with a valid named color-scale.
        See `ridgeplot.named_colorscales()` to see which names are available,
        <https://plotly.com/python/builtin-colorscales/> for more on Plotly's
        built-in color-scales.
      colormode:
        This argument controls the logic for choosing the color filling of each
        ridgeline trace. Each option provides a different method for
        calculating the `colorscale` midpoint of each trace. The default is
        mode is "mean-means". Choices are:
          - "index" - uses the trace's index. e.g. if 3 traces are specified,
           then the midpoints will be [0, 0.5, 1].
          - "mean-minmax" - uses the min-max normalized (weighted) mean of
          each density to calculate the midpoints. The normalization min and
          max values are the minimum and maximum x-values from all densities,
           respectively.
          - "mean-means" - uses the min-max normalized (weighted) mean of
          each density to calculate the midpoints. The normalization min and
          max values are the minimum and maximum mean values from all
          densities, respectively.
      coloralpha:
        If None (default), this argument will be ignored and the transparency
        values of the specifies color-scale will remain untouched. Otherwise,
        if a `float` value is passed, it will be used to overwrite the
        transparency (alpha) of the color-scale's colors.
      labels:
        A list of string labels for each trace. The default value is `None`,
        which will result in auto-generated labels of form "Trace n". If,
        instead, a list of labels is specified, it must be of the same
        size/length as the number of traces.
      linewidth:
        The traces' line width (in px).
      spacing:
        The vertical spacing between density traces, which is defined in units
        of the highest distribution (i.e. the maximum y-value).
      show_annotations:
        If `True` (default), it will show the label names as "y-tick-labels".
      xpad:
        Specifies the extra padding to use on the x-axis. It is defined in
        units of the range between the minimum and maximum x-values from all
        distributions.

    Returns:
        A Plotly figure with a ridgeline plot. You can further customize this
        figure to your liking (e.g. using the `fig.update_layout()` method).

    Raises:
      ValueError:
        If both `samples` and `densities` arguments are not `None`, or if
        neither `samples` or `densities` are specified.
    """
    has_samples = samples is not None
    has_densities = densities is not None
    if has_samples and has_densities:
        raise ValueError("You may not specify both `samples` and `densities` arguments!")
    elif not has_samples and not has_densities:
        raise ValueError("You have to specify one of: `samples` or `densities`")
    elif not has_densities:
        densities = get_densities(samples, points=kde_points, kernel=kernel, bandwidth=bandwidth)

    ridgeplot_figure_factory = _RidgePlotFigureFactory(
        densities=densities,
        labels=labels,
        colorscale=colorscale,
        coloralpha=coloralpha,
        colormode=colormode,
        linewidth=linewidth,
        spacing=spacing,
        show_annotations=show_annotations,
        xpad=xpad,
    )
    fig = ridgeplot_figure_factory.make_figure()
    return fig
